; $Id: opan__define.pro,v 1.2 2002/09/20 19:16:50 dimeo Exp $
;+
; NAME:
;       OPAN__DEFINE
;
; PURPOSE:
;
;       Object implementation of the program formerly known as PAN (Peak ANalysis).
;		This is a prototype implementation!
;
;
; AUTHOR:
;
;       Robert M. Dimeo, Ph.D.
;		NIST Center for Neutron Research
;       100 Bureau Drive
;		Gaithersburg, MD 20899
;       Phone: (301) 975-8135
;       E-mail: robert.dimeo@nist.gov
;       http://www.ncnr.nist.gov/staff/dimeo
;
; CATEGORY:
;
;       Objects, widgets, curve fitting
;
; CALLING SEQUENCE:
;
;       object = obj_new('OPAN')
;
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opanCleanup,tlb
widget_control,tlb,get_uvalue = self
if obj_valid(self) then begin
  s = size(self->getNotifyIds())
  if s[0] ne 0 then begin
    if s[0] eq 1 then count = 0 else count = s[2]-1
    for j = 0,count do begin
      OPANInfo = {OPANEvent,$
                            ID:(self->getNotifyIds())[0,j],$
                            Top:(self->getNotifyIds())[1,j],$
                            Handler:0l, $
                            object:self}
      if widget_info((self->getNotifyIds())[0,j],/valid_id) then begin $
        widget_control,(self->getNotifyIds())[0,j],send_event = OPANInfo
      endif
    endfor
  endif else begin
    obj_destroy,self
  endelse
endif
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::quit,event = event
widget_control,self.tlb,/destroy
help,/heap,/brief	; use for debugging
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::cleanup
tvlct,*self.rPtr,*self.gPtr,*self.bPtr
ptr_free,self.xtabPtr,self.btabPtr,self.ttabPtr,self.ytabPtr
ptr_free,self.resPtr,self.oneResPtr
ptr_free,self.rPtr,self.gPtr,self.bPtr
ptr_free,self.curvenamesPtr, self.headerPtr
ptr_free,self.davePtr,self.ktablePtr
ptr_free,self.dataPtr,self.errorPtr
ptr_free,self.xvalsPtr,self.yvalsPtr
ptr_free,self.datxPtr,self.datyPtr
ptr_free,self.resxPtr,self.resyPtr
ptr_free,self.ocurrentPtr,self.prefs
ptr_free,self.grpArrayPtr,self.goodParmPtr
ptr_free,self.notifyIdPtr
wdelete,self.resPix,self.datPix
; Since we have a pointer to an array of object containers, destroy
; the object container array first, then free the pointer.
if total(obj_valid(*self.ocurveGroup) ge 1) then obj_destroy,*self.ocurveGroup
ptr_free,self.ocurveGroup
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro fitfunction,xlim,parms,yfit,_Extra = extra
; This is the function that can be used in curve fitting (using Markwardt's
; routines for instance).
x = *extra.xPtr
oc = extra.oc
oc->setparms,parms
resIndex = where(tag_names(extra) eq 'RESPTR',count)

if count gt 0 then begin
  rlimit = extra.rlimit
  oc->evaluate,x,resPtr = extra.resPtr,yout = yout,rlimit = rlimit
endif else begin
  oc->evaluate,x,yout = yout,_Extra = extra
endelse

wherevalid = where(x ge min(xlim) and x le max(xlim),newpts)
yfit = yout[wherevalid]
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::initCleanup, event = event
self->clearAllCurves
if n_elements(*self.dataPtr) gt 0 then begin
  if n_elements(*self.ocurveGroup) gt 0 then begin
     obj_destroy,(*self.ocurveGroup)
  endif
endif
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::initNewDataObjects, event = event
if n_elements(*self.dataPtr) eq 0 then return
datSize = size(*self.dataPtr)
if datSize[0] eq 2 then $
  nq = (size(*self.dataPtr))[2] $
else $
  nq = 1

*self.ocurveGroup = objarr(nq)
for i = 0,nq-1 do begin
 (*self.ocurveGroup)[i] = obj_new("FUNC_CONT")
endfor
grpString = strtrim(string(1),2)+'-'+strtrim(string(nq),2)
widget_control,self.groupField,set_value = grpString
widget_control,self.groupSlider,set_slider_min = 1,set_slider_max = nq
x = (*self.xvalsPtr)
(*self.prefs).xfitlo = min(x)
(*self.prefs).xfithi = max(x)

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::clearAllCurves, event = event
; We can't just remove the objects from the container.  We actually have to
; free up each of the objects here...
print,ptr_valid(self.oneResPtr)
ngroups = n_elements(*self.ocurveGroup)
if total(obj_valid(*self.ocurveGroup)) eq 0 then return
for j = 0,ngroups-1 do begin
  ncurves = (*self.ocurveGroup)[j]->count()
  if ncurves ne 0 then begin
    oall = (*self.ocurveGroup)[j]->get(/all)
    obj_destroy,oall
    (*self.ocurveGroup)[j]->remove,/all
  endif
endfor

print,ptr_valid(self.oneResPtr)

wset,self.resPix
erase
wset,self.resVis
erase
wset,self.datPix
self->displaydata
wset,self.datVis
device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]

widget_control,self.curveSlider,set_slider_max = 2

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::clearCurrentCurves, event = event
; Clears the curves from those in the current slider position
widget_control,self.groupSlider,get_value = val
val = fix(val[0]) - 1
if n_elements(*self.dataPtr) eq 0 then return
oc = (*self.ocurveGroup)[val]
ncurves = oc->count()
if ncurves eq 0 then return

; We can't just remove the objects from the container.  We actually have to
; free up each of the objects here...
oall = oc->get(/all)
obj_destroy,oall
oc->remove,/all

wset,self.resVis
erase
wset,self.datPix
self->displaydata
wset,self.datVis
device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
widget_control,self.curveSlider,set_slider_max = 2

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::selectCurve, event = event

; Here I have used an unsatisfactory solution and implemented a modal
; widget dialog for parameter entry when the preferences are set
; such that the user can input only via a dialog or the type of
; curve demands it (i.e. it cannot be drawn with two clicks of the
; mouse or it is a user function).
ncurvetypes = n_elements(*self.curvenamesPtr)

; We begin by considering a user-defined function
if event.value eq (ncurvetypes-1) then begin
  if n_elements(*self.dataPtr) eq 0 then return

  out = enterfunction(group_leader = event.top)
  if out.cancel eq 1 then return

  parmStr = getparmsfromexpr(out.expr)
  help,parmStr,output = output
  test = strpos(output,'STRING')
  if test[0] eq (-1L) then return

  nparms = n_elements(parmStr)
  parms = replicate(1d,nparms)
  name = 'pan_userfunction'

  ; Which group are we looking at?
  widget_control,self.groupSlider,get_value = val
  val = fix(val[0])-1
  dat = (*self.dataPtr)[*,val]
  daterr = (*self.errorPtr)[*,val]
  x = (*self.xvalsPtr)[*,val]
  y = (*self.yvalsPtr)[0,val]
  ; Call the function here and see if it evaluates
  ; properly!

  test = pan_userfunction(x,parms,$
                    	  expr = out.expr, $
                    	  eval = eval)
  if eval ne 1 then return
  *self.ocurrentPtr = obj_new("func",name = name, $
                      xvalues = x, $
                      expr = out.expr, $
                      parms = parms,$
                      xtabPtr = self.xtabPtr, $
                      ytabPtr = self.ytabPtr, $
                      btabPtr = self.btabPtr, $
                      ttabPtr = self.ttabPtr)

  (*self.ocurveGroup)[val] -> add,*self.ocurrentPtr
  parmInfo = self->packageParmInfo()
endif else begin		; Now consider any of the library of functions

	name = strlowcase((*self.curvenamesPtr)[event.value])
	if name eq 'delta' and n_elements(*self.resPtr) eq 0 then begin
	  strout = 'You must load a resolution function first!'
	  void = dialog_message(dialog_parent = event.top,strout)
	  return
	endif
	name = strtrim('pan_'+name,2)
	if n_elements(*self.dataPtr) eq 0 then return

	; Which group are we looking at?
	widget_control,self.groupSlider,get_value = val
	val = fix(val[0])-1
	dat = (*self.dataPtr)[*,val]
	daterr = (*self.errorPtr)[*,val]
	x = (*self.xvalsPtr)[*,val]
	y = (*self.yvalsPtr)[0,val]

    if (n_elements(*self.resPtr) gt 0) then begin
      *self.oneResPtr = (reform((*self.resPtr).data[*,val]))
      rlimit = (*self.resPtr).rlimit

	  ; Select parameters
	  *self.ocurrentPtr = obj_new("func",name = name,xvalues = x,$
	                      xtabPtr = self.xtabPtr, $
	                      ytabPtr = self.ytabPtr, $
	                      ttabPtr = self.ttabPtr, $
   	                      btabPtr = self.btabPtr, $
   	                      resPtr = self.oneResPtr, $
   	                      rlimit = rlimit)
    endif else begin
	  *self.ocurrentPtr = obj_new("func",name = name,xvalues = x,$
	                      xtabPtr = self.xtabPtr, $
	                      ytabPtr = self.ytabPtr, $
	                      ttabPtr = self.ttabPtr, $
   	                      btabPtr = self.btabPtr)
    endelse
endelse

; Ok...now do we add the curve using a mouse or using a dialog?
initGuesses = (*self.prefs).initGuesses
if initGuesses eq 1 then begin
; Use dialog input
  (*self.ocurveGroup)[val] -> add,*self.ocurrentPtr
  parmInfo = self->packageParmInfo()

  enterparminfo,  parmInfo, $
                  group_leader = event.top, $
                  notifyIds = [event.id,event.top], $
                  modal = 1, $
                  parmInfoOut = parmInfoOut

  dummy = self->updateParmInfo(parmInfoOut)

  wset,self.resPix
  self->plotResiduals
  wset,self.resVis
  device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.resPix]

  wset,self.datPix
  self->displaydata
  wset,self.datVis
  device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]

  return
endif

*self.ocurrentPtr->getProperty,canDraw = canDraw

if (initGuesses eq 0) and (canDraw eq 1) then begin
; Use mouse input
  self.addcurve = 1
  drawMessage = *self.ocurrentPtr->getDrawMessage()
  widget_control,self.info,set_value = drawMessage[0:1]
  return
endif

if (initGuesses eq 0) and (canDraw eq 0) then begin
; Use dialog input
  parmInfo = self->packageParmInfo()
  enterparminfo,  parmInfo, $
                  group_leader = event.top, $
                  notifyIds = [event.id,event.top], $
                  modal = 1, $
                  parmInfoOut = parmInfoOut

  dummy = self->updateParmInfo(parmInfoOut)

  wset,self.resPix
  self->plotResiduals
  wset,self.resVis
  device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.resPix]

  wset,self.datPix
  self->displaydata
  wset,self.datVis
  device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
  return
endif

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::zoomEvents, event = event
case event.type of
0:	begin		; button press
	  self.mouse = event.press
	  if self.mouse eq 4 then begin
	    self.autoscale = 1

	    !x = *self.datxPtr
	    !y = *self.datyPtr
	    wset,self.datPix
	    self->displaydata
	    wset,self.datVis
	    device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
		wset,self.resPix
		self->plotresiduals
	    device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.resPix]
	  endif
	  if self.mouse eq 1 then begin
	    self.xbox[0] = event.x
	    self.ybox[0] = event.y
	    !x = *self.datxPtr
	    !y = *self.datyPtr
	    wset,self.datVis
	    device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
	    wset,self.resPix
	    self->plotresiduals
	    wset,self.resVis
	    device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.resPix]
	    empty
	    self.autoscale = 0
	    widget_control,self.datWin,/draw_motion_events
	  endif
	end
1:	begin	; button release
	 if self.mouse eq 1 then begin
	  xll = self.xbox[0] < self.xbox[1]
	  yll = self.ybox[0] < self.ybox[1]
	  w = abs(self.xbox[1] - self.xbox[0])
	  h = abs(self.ybox[1] - self.ybox[0])
	  xur = xll + w
	  yur = yll + h
	  !x = *self.datxPtr & !y = *self.datyPtr
	  ll = convert_coord(xll,yll,/device,/to_data)
	  ur = convert_coord(xur,yur,/device,/to_data)
	  self.xrange = [ll[0],ur[0]]
	  self.yrange = [ll[1],ur[1]]
	  !x = *self.datxPtr
	  !y = *self.datyPtr
	  wset,self.datPix
	  self->displaydata
	  wset,self.datVis
	  device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
	  wset,self.resPix
	  self->plotresiduals
	  wset,self.resVis
	  device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.resPix]
	  self.mouse = 0B
	  widget_control,self.datWin,draw_motion_events = 0
	 endif
	 if self.mouse eq 4 then begin
	  wset,self.datPix
	  self->displaydata
	  wset,self.datVis
	  device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
	  wset,self.resPix
	  self->plotresiduals
	  wset,self.resVis
	  device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.resPix]
	  self.mouse = 0B
	  widget_control,self.datWin,draw_motion_events = 0
	 endif
	end
2:	begin	; mouse motion
	  if self.mouse eq 1 then begin
	  	self.xbox[1] = event.x
	  	self.ybox[1] = event.y
	  	xc = [self.xbox[0],event.x,event.x,$
	  	      self.xbox[0],$
	  	      self.xbox[0]]
	  	yc = [self.ybox[0],self.ybox[0],$
	  	      event.y,event.y,$
	  	      self.ybox[0]]
		;wset,self.resPix
		;self->plotresiduals
		;wset,self.resVis
		;device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.resPix]
		!x = *self.datxPtr & !y = *self.datyPtr
	  	wset,self.datVis
	  	device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
	  	plots,xc,yc,/device


	  	empty
	  endif
	end
else:
endcase
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::drawCurveEvents, event = event
widget_control,self.groupSlider,get_value = val
val = fix(val[0])-1
xrange = self.xrange
if n_elements(*self.resPtr) gt 0 then $
  *self.oneResPtr = reform((*self.resPtr).data[*,val])

case event.type of
0:	begin		; button press
	  self.mouse = event.press
	  if self.mouse eq 1 then begin
        dcoords = convert_coord(event.x,event.y,/device,/to_data)

	    if self.addcurve eq 1 then begin
	      (*self.ocurveGroup)[val] -> add,*self.ocurrentPtr
          *self.ocurrentPtr->changefirst,dcoords[0],dcoords[1],xrange
          widget_control,self.datWin,/draw_motion_events
	    endif
	    if self.addcurve eq 2 then begin
          *self.ocurrentPtr->changesecond,dcoords[0],dcoords[1],xrange
          widget_control,self.datWin,/draw_motion_events
	    endif
        wset,self.datVis
	    device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]

		(*self.ocurveGroup)[val]->draw,overplot = 1,resPtr = self.oneResPtr
	  endif
	end
1:	begin	; button release
		dcoords = convert_coord(event.x,event.y,/device,/to_data)
		if self.mouse eq 1 then begin
		  if self.addcurve eq 1 then begin
		    *self.ocurrentPtr->changefirst,dcoords[0],dcoords[1],xrange
			widget_control,self.datWin,draw_motion_events = 0
			drawMessage = *self.ocurrentPtr->getDrawMessage()
			widget_control,self.info,set_value = drawMessage[2:3]
	      endif
		  if self.addcurve eq 2 then begin
		    *self.ocurrentPtr->changesecond,dcoords[0],dcoords[1],xrange
		    widget_control,self.datWin,draw_motion_events = 0
		    widget_control,self.info,set_value = ''
	      endif
        endif
        wset,self.datVis
	    device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
        (*self.ocurveGroup)[val]->draw,overplot = 1,resPtr = self.oneResPtr
	  self.mouse = 0B
	  self.addcurve = self.addcurve + 1
	  if self.addcurve eq 3 then begin
	    self.addcurve = 0
	    wset,self.datPix
		self->displaydata
		wset,self.datVis
		device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
		oc = (*self.ocurveGroup)[val]
		ncurves = oc->count()
	    widget_control,self.curveSlider,set_slider_max = (ncurves > 2)
	  endif
	end
2:	begin	; mouse motion
		dcoords = convert_coord(event.x,event.y,/device,/to_data)
		if self.mouse eq 1 then begin
		  if self.addcurve eq 1 then begin
		    *self.ocurrentPtr->changefirst,dcoords[0],dcoords[1],xrange
	      endif
		  if self.addcurve eq 2 then begin
		    *self.ocurrentPtr->changesecond,dcoords[0],dcoords[1],xrange
	      endif
        endif
        wset,self.datVis
	    device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
        (*self.ocurveGroup)[val]->draw,overplot = 1,resPtr = self.oneResPtr
	end
else:
endcase
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::drawEvents, event = event
if n_elements(*self.dataPtr) eq 0 then return
!x = *self.datxPtr
!y = *self.datyPtr
case self.addcurve of
0:	begin
	  self->zoomEvents,event = event
	end
1:	begin
	  self->drawCurveEvents, event = event
	end
2:	begin
	  self->drawCurveEvents, event = event
	end
else:
endcase

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::resize,event = event
ctrlgeom = widget_info(self.ctrlbase,/geometry)
infogeom = widget_info(self.infobase,/geometry)
xsize = event.x
ysize = event.y

; New data window dimensions
newxsize = xsize-ctrlgeom.xsize-infogeom.xsize
newysize = fix(ysize*(self.winRatio/(1.0+self.winRatio)))

widget_control,self.datWin,draw_xsize = newxsize, $
               draw_ysize = newysize
wdelete,self.datPix
window,/free,/pixmap,xsize = newxsize,ysize = newysize
self.datPix = !d.window

; New residual window dimensions
newysize = fix(ysize/(1.0+self.winRatio))

widget_control,self.resWin,draw_xsize = newxsize, $
               draw_ysize = newysize
wdelete,self.resPix
window,/free,/pixmap,xsize = newxsize,ysize = newysize
self.resPix = !d.window

wset,self.resPix
self->plotResiduals
wset,self.resVis
device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.resPix]

wset,self.datPix
self->displaydata
wset,self.datVis
device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opanEvents,event
if tag_names(event,/structure_name) eq 'WIDGET_BASE' then begin
  widget_control,event.top,get_uvalue = self
  self->resize,event = event
  return
endif
widget_control,event.id,get_uvalue = cmd
call_method,cmd.method,cmd.object,event = event
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::displayFitParameters, event = event
widget_control,self.groupSlider,get_value = val
val = fix(val[0])-1
if n_elements(*self.dataPtr) eq 0 then return
oc = (*self.ocurveGroup)[val]
oc->displayparms,output = output
if n_elements(output) ne 0 then begin
	widget_control,self.info,set_value = output
endif else begin
	widget_control,self.info,set_value = 'No curves selected'
endelse
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::displaydata,_Extra = extra
widget_control,self.groupSlider,get_value = val
val = fix(val[0])-1
if n_elements(*self.dataPtr) eq 0 then return
dat = (*self.dataPtr)[*,val]
daterr = (*self.errorPtr)[*,val]
x = (*self.xvalsPtr)[*,val]
y = (*self.yvalsPtr)[0,val]
thisFormat = '(f5.2)'
self.title = '!6Q='+strtrim(string(y,format = thisformat),2)

if self.autoscale eq 1 then begin
  dx = 0.1*(max(x)-min(x))
  self.xrange = [min(x)-dx,max(x)+dx]
  dy = 0.1*(max(dat)-min(dat))
  self.yrange = [min(dat)-dy,max(dat)+dy]
endif

if (*self.prefs).xenforce eq 1 then begin
  self.xrange = [(*self.prefs).xmin,(*self.prefs).xmax]
endif
if (*self.prefs).yenforce eq 1 then begin
  self.yrange = [(*self.prefs).ymin,(*self.prefs).ymax]
endif

plot,x,dat,psym = 4,xtitle = self.xtitle,ytitle = self.ztitle,title = self.title, $
     xrange = self.xrange,yrange = self.yrange,xstyle = 1,ystyle = 1,_Extra = extra
errplot,x,dat-daterr,dat+daterr,width = 0.0

if obj_valid((*self.ocurveGroup)[val]) then begin
  ncurves = (*self.ocurveGroup)[val]->count()
  ocurves = (*self.ocurveGroup)[val]
  ocurves->drawcomponents,linestyle = 2,thick = 1.5
  if n_elements(*self.resPtr) gt 0 then begin
    *self.oneResPtr = reform((*self.resPtr).data[*,val])
    ocurves->evaluate,x,yout = yout,resPtr = self.oneResPtr
  endif else begin
    ocurves->evaluate,x,yout = yout
  endelse

  if n_elements(yout) gt 1 then begin
    ; Is there a resolution function present?  Yes -> do the convolution!

    if n_elements(*self.resPtr) gt 0 then begin
      *self.oneResPtr = (reform((*self.resPtr).data[*,val]))
      rlimit = (*self.resPtr).rlimit
      ocurves->evaluate,x,yout = yout,resPtr = self.oneResPtr,rlimit = rlimit
      oplot,x,yout,linestyle = 0,thick = 2.0
    endif else begin
      oplot,x,yout,linestyle = 0,thick = 2.0
    endelse

  endif else begin
    if !d.name ne 'PS' then wset,self.resVis
    erase
  endelse

endif
*self.datxPtr = !x
*self.datyPtr = !y

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::plotResiduals, event = event,_Extra = extra
if n_elements(*self.dataPtr) eq 0 then return
widget_control,self.groupSlider,get_value = val
val = fix(val[0])-1
if not obj_valid((*self.ocurveGroup)[val]) then return

ncurves = (*self.ocurveGroup)[val]->count()
if ncurves eq 0 then begin
  erase
  return
endif
dat = (*self.dataPtr)[*,val]
daterr = (*self.errorPtr)[*,val]
x = (*self.xvalsPtr)[*,val]

if n_elements(*self.resPtr) gt 0 then begin
  *self.oneResPtr = (reform((*self.resPtr).data[*,val]))
  rlimit = (*self.resPtr).rlimit
  (*self.ocurveGroup)[val]->evaluate,x,yout = yout,resPtr = self.oneResPtr,rlimit = rlimit
endif else begin
   (*self.ocurveGroup)[val]->evaluate,x,yout = yout
endelse

residuals = (yout-dat)/daterr
plot,x,residuals,psym = 0,xrange = self.xrange,xstyle = 1,xtitle = self.xtitle,$
     ytitle = '!6Residuals',_Extra = extra
plots,!x.crange,[1.0,1.0],/data,linestyle = 2,thick = 1.5
plots,!x.crange,[-1.0,-1.0],/data,linestyle = 2,thick = 1.5
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::selectGroup, event = event
wset,self.resPix
self->plotResiduals
wset,self.resVis
device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.resPix]
wset,self.datPix
self->displaydata
wset,self.datVis
device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]

widget_control,self.groupSlider,get_value = val
val = fix(val[0])-1
if n_elements(*self.dataPtr) eq 0 then return
oc = (*self.ocurveGroup)[val]
oc->displayparms,output = output
ncurves = oc->count()
if ncurves gt 0 then begin
  widget_control,self.curveSlider,set_slider_max = (ncurves > 2)
endif
if n_elements(output) eq 0 then begin
  output = 'No curves present'
endif
widget_control,self.info,set_value = output
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
function opan_replaceSymbol,inText
outText = inText
invAngstromSym = '!3!sA!r!u!9 %!3!n!E-1!N'
omegaSym = '!4x!3'
if strpos(inText,'A-1') ne -1 then outText = invAngstromSym
if strpos(inText,'ueV') ne -1 then outText = '!4l!6eV'
if strpos(inText,'omega') ne -1 then outText = omegaSym
return,outText
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::stuffDave, event = event,inFromDave = inFromDave
; This method stuffs the DAVE pointer into the PAN data format
if n_elements(inFromDave) eq 0 then inFromDave = 0

if inFromDave eq 0 then begin
  restore,self.filename	; Restores DAVEPTR
  state = *davePtr
  ok = where(strupcase(tag_names(state)) eq 'EXTRAPTR',count)
  if count gt 0 then begin
    obj_destroy,(*state.extraPtr).paninfo.panContainer
    ptr_free,state.extraPtr
  endif
endif else begin
  state = *self.davePtr
endelse


instrument = (*state.dataStrPtr).commonStr.instrument
energy = (*(*state.dataStrPtr).commonStr.histPtr).x
xtype = (*state.dataStrPtr).commonStr.xtype
q = (*(*state.dataStrPtr).commonStr.histPtr).y
ytype = (*state.dataStrPtr).commonStr.ytype
sx = size(energy)
sy = size(q)
if (sx[0] eq 1) and (sy[0] eq 1) then begin
  if strupcase(xtype[0]) eq 'HISTOGRAM' then begin
      nx = n_elements(energy)
      ;xaverage = fltarr(nx-1)
      xaverage = 0.5*(energy[1:nx-1]+energy[0:nx-2])
      energy = xaverage
  endif
  if strupcase(ytype[0]) eq 'HISTOGRAM' then begin
    ny = n_elements(q)
    ;qaverage = fltarr(ny-1)
    qaverage = 0.5*(q[1:ny-1]+q[0:ny-2])
    q = qaverage
  endif
  nx = n_elements(energy)
  nq = n_elements(q)

  uq = 1+bytarr(nq)
  ux = 1+bytarr(nx)
  xmat = energy#uq
  ymat = ux#q
endif else begin
  xmat = energy
  ymat = q
endelse



*self.xvalsPtr = xmat
*self.yvalsPtr = ymat

data = (*(*state.dataStrPtr).commonStr.histPtr).qty
self.ztitle = (*state.dataStrPtr).commonStr.histLabel

dataSize = size(data)
*self.dataPtr = data
commonStr = (*state.dataStrPtr).commonStr
;print,tag_names(commonStr)
;help,commonStr.treatmentPtr

if n_elements(*commonStr.treatmentPtr) ne 0 then begin
  *self.headerPtr = *commonStr.treatmentPtr
endif else begin
  *self.headerPtr = 'No information available'
endelse
error = (*commonStr.histPtr).err
*self.errorPtr = error

;treatment = *(*state.dataStrPtr).commonStr.treatmentPtr
xlabel = commonStr.xlabel
xunits = commonStr.xunits
yunits = commonStr.yunits
ylabel = commonStr.ylabel

xxunits = strtrim(strmid(xunits,strpos(xunits,':')+1))
yyunits = strtrim(strmid(yunits,strpos(yunits,':')+1))
xunits = opan_replaceSymbol(xxunits)
yunits = opan_replaceSymbol(yyunits)

xlabelDisp = xlabel+'('+xunits+')'
ylabelDisp = ylabel+'('+yunits+')'

self.xtitle = xlabelDisp
self.ytitle = ylabelDisp
self.title = ylabel

if inFromDave eq 0 then $
  dave_beast_ops,"rem",davePtr,show_commands = 0,show_analysis = 0

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::scaleKtable, event = event
ktable = *self.ktablePtr
xmax = max(*self.xvalsPtr)
xmin = min(*self.xvalsPtr)
bigx = max([abs(xmax),abs(xmin)]) ; the maximum value in the dynamic range;
ksize = size(ktable)
npts = ksize[1]-1
nb = ksize[2]-1
nt = ksize[3]-1
ytab = dblarr(npts,nb,nt)
xtab = dblarr(npts)
btab = dblarr(nb)
ttab = dblarr(nt)
ytab[0:npts-1,0:nb-1,0:nt-1] = ktable[0:npts-1,0:nb-1,0:nt-1]
xtab[0:npts-1] = ktable[0:npts-1,nb,nt]
btab[0:nb-1] = ktable[npts,0:nb-1,nt]
ttab[0:nt-1] = ktable[npts,nb,0:nt-1]

xtabmax = max([abs(min(xtab)),abs(max(xtab))])
; calculate the scaling factor
lambda = xtabmax/bigx  ; scaling factor
; scale the tabulated values
xtab[0:npts-1] = (xtab[0:npts-1]/lambda)
ytab[0:npts-1,0:nb-1,0:nt-1] = lambda*ktable[0:npts-1,0:nb-1,0:nt-1]
ttab[0:nt-1] = lambda*ktable[npts,nb,0:nt-1]
*self.xtabPtr = xtab
*self.btabPtr = btab
*self.ttabPtr = ttab
*self.ytabPtr = ytab
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::clearDaveRes, event = event
; Free the resolution function pointer
if n_elements(*self.resPtr) eq 0 then return
ptr_free,self.resPtr,self.oneResPtr
self.resPtr = ptr_new(/allocate_heap)
self.oneResPtr = ptr_new(/allocate_heap)
; Now remove all of the delta functions from the function
; container
ngroups = n_elements(*self.ocurveGroup)
for i = 0,ngroups-1 do begin
  oc = (*self.ocurveGroup)[i]
  names = oc->getnames()
  isDelta = where(names eq 'pan_delta',countDelta)
  if countDelta gt 0 then begin
    oall = oc->get(/all)
    for j = 0,countDelta-1 do begin
      oDelta = oall[isDelta[j]]
      oc->remove,oDelta
      obj_destroy,oDelta
	  ncurves = oc->count()
	  widget_control,self.curveSlider,set_slider_max = (ncurves > 2)
    endfor
  endif
endfor

wset,self.datPix
self->displaydata
wset,self.datVis
device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::loadDaveRes, event = event
thisPath = self.workDir
filename = DIALOG_PICKFILE(dialog_parent = event.top,$
                           title = 'Select DAVE resolution file to restore',$
                           /read,filter = '*.dave',$
                           path = thisPath)
if filename eq '' then return

restore,filename	; Restores DAVEPTR
state = *davePtr
ok = where(strupcase(tag_names(state)) eq 'EXTRAPTR',count)
if count gt 0 then begin
    obj_destroy,(*state.extraPtr).paninfo.panContainer
    ptr_free,state.extraPtr
endif

x = (*(*state.dataStrPtr).commonStr.histPtr).x
xtype = (*state.dataStrPtr).commonStr.xtype
y = (*(*state.dataStrPtr).commonStr.histPtr).y
ytype = (*state.dataStrPtr).commonStr.ytype
sx = size(x)
sy = size(y)
if (sx[0] eq 1) and (sy[0] eq 1) then begin
  if strupcase(xtype[0]) eq 'HISTOGRAM' then begin
      nx = n_elements(x)
      xaverage = 0.5*(x[1:nx-1]+x[0:nx-2])
      x = xaverage
  endif
  if strupcase(ytype[0]) eq 'HISTOGRAM' then begin
    ny = n_elements(y)
    yaverage = 0.5*(y[1:ny-1]+y[0:ny-2])
    y = yaverage
  endif
  nx = n_elements(x)
  ny = n_elements(y)

  uy = 1+bytarr(ny)
  ux = 1+bytarr(nx)
  xmat = x#uy
  ymat = ux#y
endif else begin
  xmat = x
  ymat = y
endelse

resdata = (*(*state.dataStrPtr).commonStr.histPtr).qty
resStr = {x:xmat,y:ymat,data:resdata,rlimit:max(xmat)}

; Is there a data file present?
if n_elements(*self.dataPtr) then begin
  ; Yes, stuff the resolution function contents into a structure
  ; and get out.
  *self.resPtr = resStr
  return
endif

self->interpSigRes,resStr,event = event

dave_beast_ops,"rem",davePtr,show_commands = 0,show_analysis = 0

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::interpSigRes, resStr,event = event
; This method interpolates the signal and resolution data to the same
; x-axis.
;
; Are there the same number of yvalues in the data file and the
; resolution function?
data = *self.dataPtr
dsize = size(*self.dataPtr)
ndgroups = dsize[2]
rsize = size(resStr.data)
nrgroups = rsize[2]
if nrgroups ne ndgroups then begin
  ; They have different sizes so tell user and get out
  output = ['Signal and resolution data have different number of groups', $
            '......returning without loading resolution function.']
  void = dialog_message(dialog_parent = event.top,output)
  return
endif

xres = resStr.x & resData = resStr.data
xdat = *self.xvalsPtr
data = *self.dataPtr
error = *self.errorPtr
nx = dsize[1]
xinterp = dblarr(nx,ndgroups)
dinterp = dblarr(nx,ndgroups)
rinterp = dblarr(nx,ndgroups)
errInterp = dblarr(nx,ndgroups)

for i = 0,ndgroups-1 do begin
  xmin = min(xdat[*,i])
  xmax = max(xdat[*,i])
  dx = (xmax-xmin)/(nx-1.0)
  x = xmin+dx*dindgen(nx)	; Here is our uniform x-grid
  xinterp[*,i] = x[*]
  dinterp[*,i] = interpol(data[*,i],xdat[*,i],x[*])
  errInterp[*,i] = interpol(error[*,i],xdat[*,i],x[*])
  rinterp[*,i] = interpol(resData[*,i],xres[*,i],x[*])
endfor

*self.dataPtr = dinterp
*self.xvalsPtr = xinterp
*self.errorPtr = errInterp

; Now launch a modal dialog to determine how much offset to
; remove.
resLimit = getResLimit(*self.xvalsPtr,group_leader = event.top)
; Now pull the resolution function down...
for i = 0,ndgroups-1 do begin
  x1 = reform(xinterp[*,i])
  i1 = where(x1 ge (-resLimit) and x1 le (reslimit),count)
  if count gt 5 then begin
    yoff = 0.5*(rinterp[i1[0],i]+rinterp[i1[count-1],i])
    rinterp[*,i] = rinterp[*,i]-yoff
  endif
endfor

newStr = {x:xinterp,y:resStr.y,data:rinterp,rlimit:reslimit}
*self.resPtr = newStr

wset,self.datPix
self->displaydata
wset,self.datVis
device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::loadDave, event = event
self->initCleanup
thisPath = self.workDir
filename = DIALOG_PICKFILE(dialog_parent = event.top,$
                                              title = 'Select data file to restore',$
                                              /read,filter = '*.dave',$
                                              path = thisPath)
if filename eq '' then return

self.filename = filename
self->stuffDave,event = event

self->initNewDataObjects
self->scaleKtable
wset,self.datPix
self->displaydata
wset,self.datVis
device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::loaddata, event = event
self->initCleanup
; To be filled in later!
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::loadtasdata, event = event
self->initCleanup
path = 'c:\dimeo\data'
filenames = dialog_pickfile(path = self.workDir,/read, $
           dialog_parent = event.top,/multiple_files)
if filenames[0] eq '' then return
dout = panReadMultTAS(filenames,group_leader = event.top)
help,dout,output = output
if (strpos(output,'LONG'))[0] ne (-1) then return

x = dout.x
if n_elements(x) eq 1 then begin
  output = 'Data must contain more than one point!'
  void = dialog_message(dialog_parent = event.top,output)
  return
endif
y = 0.0
data = dout.y
error = sqrt(dout.y)

*self.dataPtr = data
*self.errorPtr = error
*self.xvalsPtr = x
*self.yvalsPtr = y
*self.headerPtr = dout.header
self.xtitle = dout.xlabel
self.ytitle = dout.ylabel
self->initNewDataObjects

wset,self.datPix
self->displaydata
wset,self.datVis
device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::loadtestdata, event = event
self->initCleanup
; Create some fake data
nchan = 170
x = rmd_makepoints(xlo = -12d,xhi = 12d,npts = nchan)
ux = 1+bytarr(nchan)

nq = 20

q = rmd_makepoints(xlo = 0.25,xhi = 1.75,npts = nq)
ro = 1.73
ao = 0.5*(1.0+beselj(q*ro,0))
uq = 1+bytarr(nq)

area = 4500.0 & bg = 0.05*area
yy = bg   + (pan_gaussian(x,[area,0d,1.0]))#ao + $
	 		(pan_gaussian(x,[5.0*area,0d,6d]))#(1.0-ao)

datmat = yy

y = dblarr(nchan,nq)
for j = 0,nq-1 do begin
  for i = 0,nchan-1 do begin
    y[i,j] = randomn(s,1,poisson = datmat[i,j])
  endfor
endfor
yerr = sqrt(y)

; Now create a matrix out of all of the data variables
xmat = x#uq
ymat = ux#q
data = y
error = yerr
bigNum = 1.e7
*self.dataPtr = data/area
*self.errorPtr = error/area
*self.xvalsPtr = xmat
*self.yvalsPtr = ymat
*self.headerPtr = ['Test data for demonstration purposes',$
                   'Simulated quasielastic scattering']
self.xtitle = '!6E (!4l!6eV)'
self.ytitle = 'Q'
self.ztitle = '!6Intensity (arb units)'
self.title = '!6Neutron Spectrum'
(*self.prefs).xlabel = self.xtitle
(*self.prefs).ylabel = self.ztitle
self->initNewDataObjects
self->scaleKtable
wset,self.datPix
self->displaydata
wset,self.datVis
device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::loadtestres, event = event
; Create some fake data
nchan = 170
x = rmd_makepoints(xlo = -12d,xhi = 12d,npts = nchan)
ux = 1+bytarr(nchan)

nq = 20

q = rmd_makepoints(xlo = 0.25,xhi = 1.75,npts = nq)
ro = 1.73
ao = 0.5*(1.0+beselj(q*ro,0))
uq = 1+bytarr(nq)

area = 4500.0 & bg = 0.05*area
;res = rmd_gaussian(x,area = 1d,center = 0d,width = 1d)
;resPtr = ptr_new(res,/no_copy)
yy = bg   + (pan_gaussian(x,[area,0d,1.0]))#ao

datmat = yy

y = dblarr(nchan,nq)
for j = 0,nq-1 do begin
  for i = 0,nchan-1 do begin
    y[i,j] = randomn(s,1,poisson = datmat[i,j])
  endfor
endfor
yerr = sqrt(y)

; Now create a matrix out of all of the data variables
xmat = x#uq
ymat = ux#q
data = y
error = yerr


resStr = {x:xmat,y:ymat,data:data}
*self.resPtr = resStr
self->interpSigRes,resStr,event = event

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::startFitCurrent, event = event
widget_control,self.groupSlider,get_value = val
val = fix(val[0])
*self.grpArrayPtr = [val]
self.status = 5
self.fitloop = 1
self->fireTimer
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::interruptFitLoop, event = event
; This will be a toggle operation only
if self.fitloop eq 1 then begin
  self.fitloop = 0
  ;widget_control,self.interrupt,set_value = 'Resume fit'
  widget_control,self.fitStatField,set_value = self.fitStatus[2]
  return
endif
if self.fitloop eq 0 then begin
  self.fitloop = 1
  ;widget_control,self.interrupt,set_value = 'Interrupt fit'
  widget_control,self.fitStatField,set_value = self.fitStatus[1]
  self->fireTimer
  return
endif
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::startFitGroups, event = event
; Which groups do we want to fit?
if n_elements(*self.dataPtr) eq 0 then return
datSize = size(*self.dataPtr)
if datSize[0] eq 1 then nq = 1 else nq = datSize[2]

widget_control,self.groupField,get_value = groupString
groupString = groupString[0]
index = opan_selectgroups(groupString)
*self.grpArrayPtr = index

; Will we be using the same parameters for all of the groups?
if (*self.prefs).same eq 1 then begin
  ; Which of the current groups has at least an initial guess?
  ; Let's grab the first one...
  oc = -1L
  goodIndex = -1
  for i = 0,nq-1 do begin
    ncurves = (*self.ocurveGroup)[i]->count()
    if goodIndex eq -1 then begin
      if ncurves gt 0 then begin
        goodIndex = i
        oc = (*self.ocurveGroup)[goodIndex]
      endif
    endif
  endfor

  nindex = n_elements(index)

  for i = 0,nindex-1 do begin
    ind = index[i]-1
    if (ind ne goodIndex) then begin

      oall = oc->get(/all)	; oc is an instance of the FUNC_CONT object class.
      						; oall is an array of FUNC object references.
      ncurves = oc->count()
	  ; Remove any existing objects in oc and destroy them
	  o_old = (*self.ocurveGroup)[ind]
	  o_old_all = o_old->get(/all)
	  nold = o_old->count()
	  if nold gt 0 then begin
	    o_old->remove,/all
		for k = 0,nold-1 do begin
			obj_destroy,o_old_all[k]
		endfor
	  endif

      for j = 0,ncurves-1 do begin

		; Now extract the curve information
		oall[j]->getproperty,name = name, $
		                     xvalues = xvalues, $
		                     step = step, $
		                     parms = parms, $
		                     fixed = fixed, $
		                     fixvalues = fixvalues, $
		                     low = low, $
		                     lovalues = lovalues, $
		                     high = high, $
		                     hivalues = hivalues, $
		                     tied = tied, $
		                     parmError = parmError, $
		                     expr = expr
		; ...and make a copy of the object
		curveCopy = obj_new('FUNC',name = name, $
		                     xvalues = xvalues, $
		                     step = step, $
		                     parms = parms, $
		                     fixed = fixed, $
		                     fixvalues = fixvalues, $
		                     low = low, $
		                     lovalues = lovalues, $
		                     high = high, $
		                     tied = tied, $
		                     hivalues = hivalues, $
		                     expr = expr,$
	                    	 xtabPtr = self.xtabPtr, $
	                    	 ytabPtr = self.ytabPtr, $
	                    	 ttabPtr = self.ttabPtr, $
	                    	 btabPtr = self.btabPtr)

        curveCopy->setproperty,parmError = parmError,/calculate
        (*self.ocurveGroup)[ind]->add,curveCopy

      endfor
    endif

  endfor

endif

self.curIndex = 0
widget_control,self.groupSlider, $
               set_value = (*self.grpArrayPtr)[self.curIndex]
self.status = 5
self.fitloop = 1
self->fireTimer
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::fitCurrent, event = event

if (self.fitloop ne 1) or (n_elements(*self.dataPtr) eq 0) then return

widget_control,self.groupSlider,get_value = val
val = fix(val[0])-1
oc = (*self.ocurveGroup)[val]

if oc->count() eq 0 then begin	; advance to the next in group if no curves
								; present in current object AND there are
								; more groups left.
  if (self.curIndex lt n_elements(*self.grpArrayPtr)-1) then begin

    self.curIndex = self.curIndex + 1
    widget_control,self.groupSlider, $
                       set_value = (*self.grpArrayPtr)[self.curIndex]
    self.fitloop = 1
    self->fireTimer
    self.status = 5

  endif
endif

if (self.status ne 1) and $
   (self.status ne 2) and $
   (self.status ne 3) and $
   (self.status ne 4) then begin
  widget_control,self.fitStatField,set_value = self.fitStatus[1]
  self->fitThisGroup

endif else begin

    if self.status eq 1 then begin
      widget_control,self.fitStatField,set_value = self.fitStatus[0]
      if (self.curIndex eq n_elements(*self.grpArrayPtr)-1) then begin

        ; all done fitting the groups
		oc = (*self.ocurveGroup)[val]
		oc->displayparms,output = output

		if n_elements(output) ne 0 then $
		   widget_control,self.info,set_value = output
		widget_control,self.interrupt,set_value = 'Interrupt fit'
        self.fitloop = 0
		wset,self.datPix
		self->displayData
		wset,self.datVis
		device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
        return

      endif

	  if (self.curIndex lt n_elements(*self.grpArrayPtr)-1) then begin
        *self.goodParmPtr = oc->getParms()
        self.curIndex = self.curIndex + 1
        widget_control,self.groupSlider, $
                       set_value = (*self.grpArrayPtr)[self.curIndex]

		widget_control,self.groupSlider,get_value = val
		val = fix(val[0])-1
		ocnew = (*self.ocurveGroup)[val]
		parms = *self.goodParmPtr
        if (*self.prefs).same eq 1 then begin
          ocnew->setparms,parms
        endif
        self.fitloop = 1
        self->fireTimer
        self.status = 5

      endif

    endif

endelse

if self.status eq 1 then begin

    wset,self.datPix
    self->displaydata
    wset,self.datVis
    device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]

    wset,self.resPix
    self->plotresiduals
    wset,self.resVis
    device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.resPix]

endif
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::fitThisGroup, event = event
widget_control,self.groupSlider,get_value = val
val = fix(val[0])-1
oc = (*self.ocurveGroup)[val]
if oc->count() ne 0 then begin

  parms = oc->getparms()
  parinfo = oc->createParInfo()

  y = (*self.dataPtr)[*,val]
  yerr = (*self.errorPtr)[*,val]
  x = (*self.xvalsPtr)[*,val]

  xPtr = ptr_new(x)
  ; If there is a resolution function present then use it!
  if n_elements(*self.resPtr) ne 0 then begin
    *self.oneResPtr = reform((*self.resPtr).data[*,val])
    rLimit = (*self.resPtr).rlimit
  endif

; Try to fit it...
    w = 1.0/yerr^2
    itmax = 1
    limits = where((x ge (*self.prefs).xfitlo) and (x le (*self.prefs).xfithi))
    xlim = x[limits]
    ylim = y[limits]
    wlim = w[limits]
    if n_elements(*self.resPtr) eq 0 then begin
      yf = mpcurvefit(xlim, ylim, wlim, parms, sigma, $
                      FUNCTION_NAME='fitfunction', $
                      status = status, $
                      /quiet, $
                      itmax = itmax, $
                      chisq = chisq, $
                      parinfo = parinfo, $
                      functargs = {oc:oc,xPtr:xPtr},$
                      /autoderivative, $
                      ERRMSG=errmsg)
    endif else begin
      yf = mpcurvefit(xlim, ylim, wlim, parms, sigma, $
                      FUNCTION_NAME='fitfunction', $
                      status = status, $
                      /quiet, $
                      itmax = itmax, $
                      chisq = chisq, $
                      parinfo = parinfo, $
                      functargs = {oc:oc,xPtr:xPtr,resPtr:self.oneResPtr, $
                                   rlimit:rlimit}, $
                      /autoderivative, $
                      ERRMSG=errmsg)
    endelse
    ptr_free,xPtr
    if n_elements(sigma) gt 0 then oc->setparmError,sigma
    self.status = status
    dof = n_elements(xlim)-n_elements(parms)
	chi_reduced = chisq/dof
    oc->setchisq,chi_reduced
    if (*self.prefs).update eq 1 then begin
      wset,self.datPix
      self->displaydata
      wset,self.datVis
      device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]

      wset,self.resPix
      self->plotresiduals
      wset,self.resVis
      device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.resPix]
    endif
    self->fireTimer
endif else begin
	wset,self.resVis
	erase
	wset,self.datPix
	self->displayData
	wset,self.datVis
	device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
endelse
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::fireTimer, event = event
widget_control,self.timerID,timer = self.duration
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::sliderCurveSel,event = event
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::remSelCurve,event = event
widget_control,self.groupSlider,get_value = val
val = fix(val[0])-1
oc = (*self.ocurveGroup)[val]
ncurves = oc->count()
if ncurves eq 0 then return
oall = oc->get(/all)

widget_control,self.curveSlider,get_value = cval
cval = fix(cval[0])
cval = cval - 1	; now it goes from 0 to ncurves-1
if cval eq ncurves then return	; too big!
oc->remove,oall[cval]
obj_destroy,oall[cval]
widget_control,self.curveSlider,set_slider_max = (cval > 2)

wset,self.resPix
self->plotResiduals
wset,self.resVis
device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.resPix]

wset,self.datPix
self->displayData
wset,self.datVis
device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::plotEISF, event = event
; No data?->Get out.
if n_elements(*self.dataPtr) eq 0 then return

; Count how many groups of curves there are.  If there are less than two
; then get out.
fitObject = *self.ocurveGroup
ngroups = n_elements(fitObject)
totalFits = 0
for i = 0,ngroups-1 do begin
  oc = fitObject[i]
  ncurves = oc->count()
  if ncurves gt 0 then begin
    totalFits = totalFits + 1
    if (n_elements(okIndex) eq 0) then $
       okIndex = i else okIndex = [okIndex,i]
  endif
endfor
if totalFits lt 2 then return

thisEvent = tag_names(event,/structure_name)
case thisEvent of

'WIDGET_BUTTON': $
  begin
    opan_plotEISF,	fitObject, $
                    *self.yvalsPtr, $
					group_leader = event.top, $
                   	notifyIds = [event.id,event.top], $
                   	xtitle = self.ytitle, $
                   	ytitle = 'Value', $
                   	title = '', $
                   	workDir = self.workDir

  end
else:
endcase

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::plotFitParameter, event = event
; No data?->Get out.
if n_elements(*self.dataPtr) eq 0 then return

; Count how many groups of curves there are.  If there are less than two
; then get out.
fitObject = *self.ocurveGroup
ngroups = n_elements(fitObject)
totalFits = 0
for i = 0,ngroups-1 do begin
  oc = fitObject[i]
  ncurves = oc->count()
  if ncurves gt 0 then begin
    totalFits = totalFits + 1
    if n_elements(okIndex) eq 0 then okIndex = i else okIndex = [okIndex,i]
  endif
endfor
if totalFits lt 2 then return

thisEvent = tag_names(event,/structure_name)
case thisEvent of

'WIDGET_BUTTON': $
  begin
    opan_plotParms,	fitObject, $
                    *self.yvalsPtr, $
					group_leader = event.top, $
                   	notifyIds = [event.id,event.top], $
                   	xtitle = self.ytitle, $
                   	ytitle = 'Value', $
                   	title = '', $
                   	workDir = self.workDir

  end
;'PLOTPARMSEVENT': $
;  begin
;  end
else:
endcase

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::saveParmsAsText,event = event
if n_elements(*self.dataPtr) eq 0 then return
thisPath = self.workDir
fitObject = *self.ocurveGroup
ngroups = n_elements(fitObject)
ncurves = 0
for i = 0,ngroups-1 do begin
  oc = fitObject[i]
  ncurves = ncurves + (oc->count())
endfor
if ncurves eq 0 then return

filename = DIALOG_PICKFILE(dialog_parent = event.top,$
                                              title = 'Enter parameter text file name',$
                                              /read,filter = '*.txt',$
                                              path = thisPath)
if filename eq '' then return
filename = filename + '.txt'
openw,lun,filename,/get_lun

for i = 0,ngroups-1 do begin
  oc = fitObject[i]
  chisq = oc->getchisq()
  if chisq ne (-1.0) then begin
    ncurves = oc->count()
    if ncurves ne 0 then begin
      printf,lun,'###############################'
      printf,lun,'Group: ',strtrim(string(i+1),2)
      printf,lun,'###############################'
      oc->displayparms,output = output
      nlines = n_elements(output)
      for j = 0,nlines-1 do begin
        printf,lun,output[j]
      endfor
      printf,lun,'###############################'
    endif
  endif
endfor
free_lun,lun
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::print2text,event = event
if n_elements(*self.dataPtr) eq 0 then return
thisPath = self.workDir
filename = DIALOG_PICKFILE(dialog_parent = event.top,$
                                              title = 'Enter output text file name',$
                                              /read,filter = '*.txt',$
                                              path = thisPath)
filename = filename + '.txt'

fitObject = *self.ocurveGroup
ngroups = n_elements(fitObject)

openw,lun,filename,/get_lun

for i = 0,ngroups-1 do begin
  oc = fitObject[i]
  oall = oc->get(/all)
  ncurves = oc->count()

  dat = reform((*self.dataPtr)[*,i])
  daterr = reform((*self.errorPtr)[*,i])
  x = reform((*self.xvalsPtr)[*,i])
  nx = n_elements(x)

  strout1 = '# Group '+string(i+1)
  strout2 = '# x y yerr'
  str = ['#','x','yerr']
  if ncurves gt 0 then begin
    oc->evaluate,x,yout = yfit
    txt = replicate('  component: ',ncurves)
    txt = txt + strtrim(string(indgen(ncurves)+1),2)
    strout2 = strout2 + ' total fit '
    str = [str,'total fit']
    ycomp = fltarr(nx,ncurves)
    for j = 0,ncurves-1 do begin
      oall[j]->setproperty,xvalues = x,/calculate
      oall[j]->getproperty,yvalues = yvalues
      ycomp[*,j] = yvalues
      if j eq 0 then begin
        out = txt[j]
      endif else begin
		out = out + txt[j]
      endelse
      str = [str,txt[j]]
    endfor
    strout2 = strout2 + out
    printf,lun,strout1
    outFormat = '(A1,'+strtrim(string(3+ncurves),2)+'A20)'
    printf,lun,strout2;,format = outFormat

    thisFormat = '('+strtrim(string(4+ncurves),2)+'f15.5)'
    dataOut = fltarr(4+ncurves)
    for j = 0,nx-1 do begin
      dataOut[0:4+ncurves-1] = [x[j],dat[j],daterr[j],yfit[j],transpose(ycomp[j,0:ncurves-1])]
      printf,lun,dataOut[0:4+ncurves-1],format = thisFormat
    endfor

  endif else begin
    printf,lun,strout1
    outFormat = '(A1,'+strtrim(string(2),2)+'A20)'
    printf,lun,strout2,format = outFormat
    thisFormat = '('+strtrim(string(3),2)+'f15.5)'
    dataOut = fltarr(3)
    for j = 0,nx-1 do begin
      dataOut[0:2] = [x[j],dat[j],daterr[j]]
      printf,lun,dataOut[0:2],format = thisFormat
    endfor
  endelse
printf,lun,''
endfor

free_lun,lun
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
function opan::packageParmInfo
; Get parameter information
widget_control,self.groupSlider,get_value = val
val = fix(val[0]) - 1
oc = (*self.ocurveGroup)[val]
ncurves = oc->count()
if ncurves eq 0 then return,(-1L)
names = oc->getnames()
ncurves = n_elements(names)
step = oc->getstep()
fixed = oc->getfixed()
lovalues = oc->getlowvalues()
hivalues = oc->gethighvalues()
parms = oc->getparms()
parmnames = oc->getparmnames()
high = oc->gethigh()
low = oc->getlow()
tied = oc->gettied()
expr = oc->getExpr()

oall = oc->get(/all)
for i = 0,ncurves-1 do begin
 	oall[i]->getproperty,parmnames = parmnames
 	np = n_elements(parmnames)
 	if i eq 0 then begin
   		pn = parmnames
   		nparms = np
 	endif else begin
   		nparms = [nparms,np]
   		pn = [pn,parmnames]
 	endelse
endfor
parmInfo = {  $
			  nparms:nparms,names:names, step:step, fixed:fixed, $
    	      lovalues:lovalues, hivalues:hivalues, high:high, $
        	  low:low, parms:parms, parmnames:pn,tied:tied,expr:expr  $
        	  }
return,parmInfo
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
function opan::updateParmInfo,newParmInfo
widget_control,self.groupSlider,get_value = val
val = fix(val[0]) - 1
oc = (*self.ocurveGroup)[val]
oc->setparms,newParmInfo.parms
oc->setlow,newParmInfo.low
oc->sethigh,newParmInfo.high
oc->setlowvalues,newParmInfo.lovalues
oc->sethighvalues,newParmInfo.hivalues
oc->setfixed,newParmInfo.fixed
oc->settied,newParmInfo.tied
return,1
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::parameterInfo, event = event
if n_elements(*self.dataPtr) eq 0 then return
; Get parameter information

thisEvent = tag_names(event,/structure_name)
case thisEvent of
'': $
  begin
    widget_control,event.top,sensitive = 0
    parmInfo = self->packageParmInfo()
	widget_control,self.modify,sensitive = 0
    enterparminfo,parmInfo, $
                  group_leader = event.top, $
                  notifyIds = [event.id,event.top]
  end
'WIDGET_BUTTON': $
  begin
    widget_control,event.top,sensitive = 0
	parmInfo = self->packageParmInfo()
	widget_control,self.modify,sensitive = 0
    enterparminfo,parmInfo, $
                  group_leader = event.top, $
                  notifyIds = [event.id,event.top]
  end
'PARMSEVENT': $
  begin
	  dummy = self->updateParmInfo(*event.parmInfo)

	  wset,self.datPix
	  self->displayData
	  wset,self.datVis
	  device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]

	  wset,self.resPix
	  self->plotResiduals
	  wset,self.resVis
	  device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.resPix]

  end
'PARMSQUITEVENT': $
  begin
	widget_control,self.modify,sensitive = 1
	widget_control,event.top,sensitive = 1
	; Empty placeholder.  We don't want to update any of the
	; parameter information.
  end
else:
endcase

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::showFileInfo,event = event
thisEvent = tag_names(event,/structure_name)
if n_elements(*self.headerPtr) eq 0 then begin
  output = ['No data currently loaded']
  void = dialog_message(dialog_parent = event.top,output)
  return
endif
case thisEvent of
'WIDGET_BUTTON': $
  begin
    opan_FileInfo,*self.headerPtr,group_leader = event.top, $
              notifyIds = [event.id,event.top]

  end
'FILEINFOEVENT': $	; This case not really necessary
  begin
  end
else:
endcase

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::preferences, event = event
thisEvent = tag_names(event,/structure_name)
case thisEvent of
'WIDGET_BUTTON': $
  begin
    enterPref,group_leader = event.top, $
              notifyIds = [event.id,event.top], $
              prefs = *self.prefs

  end
'PREFSEVENT': $
  begin
    *self.prefs = event.prefs
    self.xtitle = (*self.prefs).xlabel
    self.ztitle = (*self.prefs).ylabel
	  wset,self.resPix
	  self->plotResiduals
	  wset,self.resVis
	  device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.resPix]

	  wset,self.datPix
	  self->displayData
	  wset,self.datVis
	  device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
  end
else:
endcase

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::createWidgets
; Widget creation module
if self.group_leader eq (-1L) then begin
  tlb = widget_base(/row, $
        title = 'PAN: Peak Analysis',mbar = bar,/tlb_size_events)
endif else begin
  tlb = widget_base(group_leader = self.group_leader,/row, $
        title = 'PAN: Peak Analysis',mbar = bar,/tlb_size_events)
endelse
self.tlb = tlb
filemenu = widget_button(bar,value = 'File',/menu)
printmenu = widget_button(bar,value = 'Print',/menu)
resmenu = widget_button(bar,value = 'Resolution',/menu)
void = widget_button(filemenu,value = 'Load DAVE file', $
       uvalue = {object:self,method:'loadDave'})
void = widget_button(filemenu,value = 'Load TAS data file', $
       uvalue = {object:self,method:'loadtasdata'})
void = widget_button(filemenu,value = 'Load test data', $
       uvalue = {object:self,method:'loadtestdata'})
void = widget_button(filemenu,value = 'Save fit parameters as text',$
       uvalue = {object:self,method:'saveParmsAsText'})
void = widget_button(filemenu,value = 'Display current file information',$
       uvalue = {object:self,method:'showFileInfo'})
void = widget_button(filemenu,value = 'Preferences',$
       uvalue = {object:self,method:'preferences'})
void = widget_button(filemenu,value = 'Quit',/separator,$
       uvalue = {object:self,method:'quit'})

void = widget_button(printmenu,value = 'Print to PS file', $
       uvalue = {object:self,method:'print2ps'})
void = widget_button(printmenu,value = 'Print data and fits to text file', $
       uvalue = {object:self,method:'print2text'})

void = widget_button(resmenu,value = 'Load DAVE data as resolution function', $
       uvalue = {object:self,method:'loadDaveRes'})
void = widget_button(resmenu,value = 'Clear resolution function', $
       uvalue = {object:self,method:'clearDaveRes'})
void = widget_button(resmenu,value = 'Load test resolution function', $
       uvalue = {object:self,method:'loadtestres'})

ctrlBase = widget_base(tlb,/col,uvalue = {object:self,method:'fitCurrent'})
self.ctrlBase = ctrlBase
self.timerID = ctrlBase

curveNames = ['Select function',$
			  'delta', $
              'gaussian',$
  			  'lorentzian',$
  			  'lognormal', $
  			  'DHO', $
  			  'kohlrausch', $
  			  'step', $
              'background', $
              'userfunction']

ncurves = n_elements(curveNames)
*self.curvenamesPtr = curveNames

desc = replicate({ flags:0, name:'' }, ncurves)
desc.flags = [1,0*indgen(ncurves-1)]
desc.name = curveNames

drop = cw_pdmenu(ctrlbase,desc,/return_index, $
       uvalue = {object:self,method:'selectCurve'})
void = widget_label(ctrlBase,value = '')
groupSlider = widget_slider(ctrlbase,value = 1,min = 1,max = 100, $
              title = 'Group selection', $
              uvalue = {object:self,method:'selectGroup'})
void = widget_button(ctrlbase,value = 'Clear all curves', $
       uvalue = {object:self,method:'clearAllCurves'})

self.modify = widget_button(ctrlbase,value = 'Modify fit parameters', $
              uvalue = {object:self,method:'parameterInfo'})
void = widget_button(ctrlbase,value = 'Clear current curves', $
       uvalue = {object:self,method:'clearCurrentCurves'})
void = widget_button(ctrlbase,value = 'Remove selected curve', $
       uvalue = {object:self,method:'remSelCurve'})
self.curveSlider = widget_slider(self.ctrlBase,value = 1, $
                   min = 1,max = 2,title = 'Curve selection', $
                   uvalue = {object:self,method:'sliderCurveSel'})

void = widget_label(ctrlBase,value = '')

void = widget_button(ctrlbase,value = 'Fit current group', $
       uvalue = {object:self,method:'startFitCurrent'})
void = widget_button(ctrlbase,value = 'Fit all groups', $
       uvalue = {object:self,method:'startFitGroups'})
self.interrupt = widget_button(ctrlbase,value = 'Interrupt fit', $
            uvalue = {object:self,method:'interruptFitLoop'})

self.groupSlider = groupSlider
self.groupField = cw_field(ctrlbase,/col,value = 1,/string, $
                  title = 'Groups to fit')
self.fitStatField = cw_field(ctrlbase,/col,value = self.fitStatus[0],$
                    title = 'Fitting status',/string,/noedit)

void = widget_button(ctrlbase,value = 'Display fit parameters', $
       uvalue = {object:self,method:'displayFitParameters'})

void = widget_button(ctrlbase,value = 'Plot fit parameter', $
       uvalue = {object:self,method:'plotFitParameter'})
void = widget_button(ctrlbase,value = 'Plot EISF', $
       uvalue = {object:self,method:'plotEISF'})


dispBase = widget_base(tlb,/col)
self.dispBase = dispBase

datxsize = 500 & datysize = 400
resxsize = 500 & resysize = 200
self.winratio = float(datysize/resysize)
self.datwin = widget_draw(dispBase,xsize = datxsize,ysize = datysize,$
              /button_events,uvalue = {object:self,method:'drawEvents'})
self.reswin = widget_draw(dispBase,xsize = resxsize,ysize = resysize)

self.infoBase = widget_base(tlb,/col)
thisFont = "Comic Sans MS*16"
self.info = widget_text(self.infoBase,font = thisFont, value = '', $
            xsize = 30,ysize = 20,/scroll)

; Center the interface in the screen
geom = widget_info(self.tlb,/geometry)
device,get_screen_size = sz
sx = sz[0] & sy = sz[1]
xoff = fix(0.5*(sx-geom.xsize))
yoff = fix(0.5*(sy-geom.ysize))
widget_control,self.tlb,xoffset = xoff,yoffset = yoff

widget_control,tlb,/realize

widget_control,self.datwin,get_value = datVis
self.datVis = datVis
widget_control,self.reswin,get_value = resVis
self.resVis = resVis

window,/free,/pixmap,xsize = datxsize,ysize = datysize
self.datPix = !d.window
window,/free,/pixmap,xsize = resxsize,ysize = resysize
self.resPix = !d.window

widget_control,tlb,set_uvalue = self
xmanager,'opan::createWidgets',self.tlb,event_handler = 'opanEvents',$
         cleanup = 'opanCleanup', /no_block

return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::niceOutput,event = event
; First let's position the plot at the top of the page
;ysize = 700
;xsize = fix(8.5*ysize/11.0)
;window,/free,xsize = xsize,ysize = ysize
deltay = 0.20
yreso = 0.25 & yresf = 0.40
ydato = 0.45 & ydatf = 0.75
pres = [0.15,yreso+deltay,0.9,yresf+deltay]
pdat = [0.15,ydato+deltay,0.9,ydatf+deltay]
self->plotresiduals,position = pres
self->displaydata,position = pdat,/noerase

; Now annotate with the fit results
widget_control,self.groupSlider,get_value = val
val = fix(val[0])-1
oc = (*self.ocurveGroup)[val]
oc->displayparms,output = output
chsize = 0.75
nlines = n_elements(output)
if nlines eq 0 then return
if nlines lt 30 then begin
  ystart = 0.05 & yend = ydato-ystart
  dy = yend-ystart
  step = dy/30.0
  xstart = 0.05
  ypos = ystart+step*findgen(nlines)
  for i = 0,nlines-1 do begin
    xyouts,xstart,yend-ypos[i],output[i],/normal,charsize = chsize
  endfor
endif
if (nlines gt 29) and (nlines lt 59) then begin
  ; Print the first 30 lines
  ystart = 0.05 & yend = ydato-ystart
  dy = yend-ystart
  step = dy/30.0
  xstart = 0.05
  ypos = ystart+step*findgen(30)
  for i = 0,29 do begin
    xyouts,xstart,yend-ypos[i],output[i],/normal,charsize = chsize
  endfor
  ; Print the remainder
  ypos = ystart+step*findgen(30)
  ystart = 0.05 & yend = ydato-ystart
  xstart = 0.50
  ypos = ystart+step*findgen(30)
  count = 0
  for i = 30,nlines-1 do begin
    xyouts,xstart,yend-ypos[count],output[i],/normal,charsize = chsize
    count = count + 1
  endfor
endif
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan::print2ps,event = event
if n_elements(*self.dataPtr) eq 0 then return
keywords = PSConfig(Cancel=cancelled,group_leader = event.top,$
           filename = 'panplot.ps',color = 0,xsize = 8.0,ysize = 11.0, $
           directory = self.workDir,xoff = 0.0,yoff = 0.0)
IF cancelled THEN RETURN
thisDevice = !D.Name
Set_Plot, 'PS'
Device, _Extra=keywords

self->niceOutput,event = event

Device, /Close_File
Set_Plot, thisDevice
return
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
function opan::getNotifyIds
return,*self.notifyIdPtr
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
function opan::init,	group_leader = group_leader, $
						notifyId = notifyId, $
						davePtr = davePtr, $
						data = data, $
						error = error, $
						xvalues = xvalues, $
						yvalues = yvalues, $
						xtitle = xtitle, $
						ytitle = ytitle, $
						ztitle = ztitle, $
						header = header, $
						title = title, $
						workDir = workDir, $
						dataDir = dataDir
tvlct,r,g,b,/get
self.rPtr = ptr_new(r)
self.gPtr = ptr_new(g)
self.bPtr = ptr_new(b)
loadct,0,/silent

if n_elements(group_leader) eq 0 then group_leader = (-1L)
if n_elements(workDir) eq 0 then self.workDir = '' else self.workDir = workDir
if n_elements(dataDir) eq 0 then self.dataDir = '' else self.dataDir = dataDir
self.resPtr = ptr_new(/allocate_heap)
self.oneResPtr = ptr_new(/allocate_heap)
self.xtabPtr = ptr_new(/allocate_heap)
self.ytabPtr = ptr_new(/allocate_heap)
self.btabPtr = ptr_new(/allocate_heap)
self.ttabPtr = ptr_new(/allocate_heap)
self.ktablePtr = ptr_new(/allocate_heap)
self.davePtr = ptr_new(/allocate_heap)
self.notifyIdPtr = ptr_new(/allocate_heap)
self.headerPtr = ptr_new(/allocate_heap)
self.dataPtr = ptr_new(/allocate_heap)
self.errorPtr = ptr_new(/allocate_heap)
self.xvalsPtr = ptr_new(/allocate_heap)
self.yvalsPtr = ptr_new(/allocate_heap)
self.resxPtr = ptr_new(/allocate_heap)
self.resyPtr = ptr_new(/allocate_heap)
self.datxPtr = ptr_new(/allocate_heap)
self.datyPtr = ptr_new(/allocate_heap)
self.goodParmPtr = ptr_new(/allocate_heap)
self.curvenamesPtr = ptr_new(/allocate_heap)
self.ocurveGroup = ptr_new(/allocate_heap)
self.grpArrayPtr = ptr_new(/allocate_heap)
self.status = 5
self.fitloop = 0
self.duration = 1.e-3
self.curIndex = 0
self.info = 0L
self.groupField = 0L
self.fitStatus = ['Resting', 'Fitting', 'Paused']
self.fitStatField = 0L


self.xtitle = 'x'
self.ytitle = 'y'
self.ztitle = 'z'

self.prefs = ptr_new({preferences, $
                      same:1, update:0, xenforce:0,yenforce:0, $
                      xmin:0.0,xmax:1.0,ymin:0.0,ymax:1.0, $
                      xfitlo:0.0,xfithi:1.0,initGuesses:0,tied:0, $
                      xlabel:self.xtitle,ylabel:self.ytitle})
self.fitgroup = 0
self.addcurve = 0
self.firstPoint = [0,0]
self.secondPoint = [0,0]
self.ocurrentPtr = ptr_new(/allocate_heap)

self.autoscale = 1
self.xbox = [0.0,1.0]
self.xrange = [0.0,1.0]
self.ybox = [0.0,1.0]
self.yrange = [0.0,1.0]

self.title = 'Data'
self.filename = ''

if n_elements(notifyId) ne 0 then *self.notifyIdPtr = notifyId
self.group_leader = group_leader

self->createWidgets

pan_readKTable,ktablePtr
widget_control,self.info,set_value = 'Lookup tables loaded'
*self.ktablePtr = *ktablePtr
ptr_free,ktablePtr

if n_elements(data) ne 0 then begin
  *self.dataPtr = data
  if n_elements(error) ne 0 then *self.errorPtr = error
  if n_elements(header) ne 0 then *self.headerPtr = header
  if n_elements(xvalues) ne 0 then *self.xvalsPtr = xvalues
  if n_elements(yvalues) ne 0 then *self.yvalsPtr = yvalues
  if n_elements(ztitle) ne 0 then begin
    self.ztitle = ztitle
    (*self.prefs).ylabel = ztitle
  endif
  if n_elements(xtitle) ne 0 then begin
    self.xtitle = xtitle
    (*self.prefs).xlabel = xtitle
  endif
  if n_elements(ytitle) ne 0 then begin
    self.ytitle = ytitle

    (*self.prefs).ylabel = ytitle
  endif
  if n_elements(title) ne 0 then self.title = title
  self->initNewDataObjects
  self->scaleKtable
  wset,self.datPix
  self->displayData
  wset,self.datVis
  device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
endif

if n_elements(davePtr) ne 0 then begin
  *self.davePtr = *davePtr
  self->stuffDave,inFromDave = 1
  self->initNewDataObjects
  self->scaleKtable
  wset,self.datPix
  self->displayData
  wset,self.datVis
  device,copy = [0,0,!d.x_size,!d.y_size,0,0,self.datPix]
endif

return,1
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan__define

define = 	{opan, $

			 dataDir:'', $
			 workDir:'', $
			 davePtr:ptr_new(), $
			 notifyIdPtr:ptr_new(), $
			 headerPtr:ptr_new(), $
			 prefs:ptr_new(), $
			 fitgroup:0, $
			 grpArrayPtr:ptr_new(), $
			 curIndex:0, $
			 goodParmPtr:ptr_new(), $
			 filename:'', $
			 xtitle:'', $
			 ytitle:'', $
			 ztitle:'', $
			 title:'', $
			 group_leader:0L, $
			 fitStatus:strarr(3), $
			 fitStatField:0L, $
			 modify:0L, $
			 groupField:0L, $
			 interrupt:0L, $
			 info:0L, $
			 curveSlider:0L, $
			 ocurveGroup:ptr_new(), $
			 ocurrentPtr:ptr_new(), $
			 addcurve:0, $
			 curvenamesPtr:ptr_new(), $
			 dataPtr:ptr_new(), $
			 errorPtr:ptr_new(), $
			 xvalsPtr:ptr_new(), $
			 yvalsPtr:ptr_new(), $
			 resPtr:ptr_new(), $
			 oneResPtr:ptr_new(), $
			 ktablePtr:ptr_new(), $
			 xtabPtr:ptr_new(), $
			 ytabPtr:ptr_new(), $
			 btabPtr:ptr_new(), $
			 ttabPtr:ptr_new(), $

			 autoscale:0, $
			 mouse:0B, $
			 xrange:fltarr(2), $
			 yrange:fltarr(2), $
			 xbox:fltarr(2), $
			 ybox:fltarr(2), $
			 firstPoint:intarr(2), $
			 secondPoint:intarr(2), $

			 datWin:0L, $
			 datVis:0L, $
			 datPix:0L, $
			 datxPtr:ptr_new(), $
			 datyPtr:ptr_new(), $

			 resWin:0L, $
			 resVis:0L, $
			 resPix:0L, $
			 resxPtr:ptr_new(), $
			 resyPtr:ptr_new(), $

			 winratio:0., $
			 groupSlider:0L, $
			 tlb:0L, $
			 ctrlbase:0L, $
			 dispBase:0L, $
			 infoBase:0L, $
			 duration:0.0, $
			 status:0, $
			 fitloop:0, $
			 rPtr:ptr_new(), $
			 gPtr:ptr_new(), $
			 bPtr:ptr_new(), $
			 timerID:0L}

end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro opan_example
o = obj_new('opan')
end